# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from networking_sfc.extensions import flowclassifier
from networking_sfc.extensions import servicegraph
from networking_sfc.extensions import sfc
from networking_sfc.services.flowclassifier.common import config as fc_config
from networking_sfc.services.sfc.common import config as sfc_config


def list_quota_opts():
    return [
        ('quotas',
         itertools.chain(
             flowclassifier.flow_classifier_quota_opts,
             sfc.sfc_quota_opts,
             servicegraph.service_graph_quota_opts)
         ),
    ]


def list_sfc_opts():
    return [
        ('flowclassifier', fc_config.FLOWCLASSIFIER_DRIVER_OPTS),
        ('sfc', sfc_config.SFC_DRIVER_OPTS),
    ]
